import pandas as pd
from pathlib import Path


def clean_value(value):
    value = value.strip()
    if value.startswith("'") and value.endswith("'"):
        value = value[1:-1]
    return value


work_dir = Path(__file__).parent
df = pd.read_csv(work_dir / "plants.csv")
for column in ["family", "genus", "species", "authority"]:
    values = df[~pd.isna(df[column])][column].unique()
    values = set(clean_value(value) for value in values)
    values = set(value for value in values if value)
    values = sorted(values)
    output_path = work_dir / f"{column}.txt"
    print(output_path)
    output_path.write_text("\n".join(values))



import warnings
from transformers import TrOCRProcessor, VisionEncoderDecoderModel
from PIL import Image
from enum import Enum
from pathlib import Path
import pytesseract


class TrOCRSize(Enum):
    SMALL = "small"
    BASE = "base"
    LARGE = "large"

    def __str__(self):
        return str(self.value)


class OCR:
    def get_text(self, image_path: Path) -> str:
        raise NotImplementedError


class TrOCR(OCR):
    def __init__(self, size: TrOCRSize = TrOCRSize.BASE):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            print("Getting TrOCRProcessor")
            self.processor = TrOCRProcessor.from_pretrained(
                f"microsoft/trocr-{size}-handwritten"
            )

            print("Getting VisionEncoderDecoderModel")
            self.model = VisionEncoderDecoderModel.from_pretrained(
                f"microsoft/trocr-{size}-handwritten"
            )

    def get_text(self, image_path: Path) -> str:
        image = Image.open(image_path)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            pixel_values = self.processor(
                images=image, return_tensors="pt"
            ).pixel_values
            generated_ids = self.model.generate(pixel_values)
            generated_text = self.processor.batch_decode(
                generated_ids, skip_special_tokens=True
            )[0]

        return generated_text


class Tesseract(OCR):
    def __init__(self):
        super().__init__()
        
        self.no_tesseract = False # This will be set to True later if it isn't found.

    def get_text(self, image_path: Path) -> str:
        if self.no_tesseract:
            return None

        try:
            return pytesseract.image_to_string(str(image_path)).strip()
        except Exception as err:
            print(f"No tesseract available: {err}")
            self.no_tesseract = True
        
        return None




from typing import List, Dict
from pathlib import Path
import pandas as pd
from functools import cached_property
import torch
from torchapp.examples.image_classifier import ImageClassifier
from rich.console import Console
from ultralytics import YOLO

from .yolo import yolo_output, predictions_filename
from .ocr import Tesseract, TrOCR, TrOCRSize
from .download import get_location
from .util import read_reference, ocr_data_df, adjust_text, get_stub
from .ocr import TrOCRSize
from .report import write_report

console = Console()

DEFAULT_RELEASE_PREFIX = "https://github.com/rbturnbull/hespi/releases/download/v0.4.0"
DEFAULT_SHEET_COMPONENT_WEIGHTS = f"{DEFAULT_RELEASE_PREFIX}/sheet-component.pt.gz"
DEFAULT_LABEL_FIELD_WEIGHTS = f"{DEFAULT_RELEASE_PREFIX}/label-field.pt.gz"
DEFAULT_INSTITUTIONAL_LABEL_CLASSIFIER_WEIGHTS = f"https://github.com/rbturnbull/hespi/releases/download/v0.1.0-alpha/institutional-label-classifier.pkl"


class Hespi():
    def __init__(
        self,
        trocr_size: TrOCRSize = TrOCRSize.BASE.value,
        sheet_component_weights: str = "",
        label_field_weights: str = "",
        institutional_label_classifier_weights: str = "",
        force_download:bool = False,
        gpu: bool = True,
        fuzzy: bool = True,
        fuzzy_cutoff: float = 0.8,
        htr: bool = True,
        tmp_dir:str = None,
        batch_size:int = 4,
        sheet_component_res:int = 1280,
        label_field_res:int = 1280,
    ):
        self.trocr_size = trocr_size
        self.sheet_component_weights = sheet_component_weights or DEFAULT_SHEET_COMPONENT_WEIGHTS
        self.label_field_weights = label_field_weights or DEFAULT_LABEL_FIELD_WEIGHTS
        self.institutional_label_classifier_weights = institutional_label_classifier_weights or DEFAULT_INSTITUTIONAL_LABEL_CLASSIFIER_WEIGHTS
        self.force_download = force_download
        self.fuzzy = fuzzy
        self.fuzzy_cutoff = fuzzy_cutoff
        self.htr = htr
        self.tmp_dir = tmp_dir
        self.batch_size = batch_size
        self.sheet_component_res = sheet_component_res
        self.label_field_res = label_field_res
        
        # Check if gpu is available
        self.gpu = gpu and torch.cuda.is_available()
        self.device = "cuda:0" if self.gpu else "cpu"

    def get_yolo(self, weights_url:str) -> YOLO:
        weights = get_location(weights_url, force=self.force_download)
        model = YOLO(weights)
        model.to(self.device)
        return model

    @cached_property
    def sheet_component_model(self):
        return self.get_yolo(self.sheet_component_weights)

    @cached_property
    def label_field_model(self):
        return self.get_yolo(self.label_field_weights)

    @cached_property
    def institutional_label_classifier(self):
        model = ImageClassifier()
        model.pretrained = get_location(self.institutional_label_classifier_weights, force=self.force_download)
        return model

    @cached_property
    def reference(self):
        reference_fields = ["family", "genus", "species", "authority"]
        return {field: read_reference(field) for field in reference_fields}

    @cached_property
    def tesseract(self):
        return Tesseract()

    @cached_property
    def trocr(self):
        print(f"Getting TrOCR model of size '{self.trocr_size}'")
        trocr = TrOCR(size=self.trocr_size)
        print("TrOCR model available")
        return trocr

    def sheet_component_detect(
        self,
        images:List[Path],
        output_dir:Path,
    ):
        return yolo_output(
            self.sheet_component_model, 
            images, 
            output_dir=output_dir, 
            tmp_dir_prefix=self.tmp_dir, 
            res=self.sheet_component_res,
            batch_size=self.batch_size,
        )

    def label_field_model_detect(
        self,
        images:List[Path],
        output_dir:Path,
    ):
        return yolo_output(
            self.label_field_model, 
            images, 
            output_dir=output_dir, 
            tmp_dir_prefix=self.tmp_dir, 
            res=self.label_field_res,
            batch_size=self.batch_size,
        )

    def detect(
        self,
        images:List[Path],
        output_dir:Path,
        report:bool = True,
    ):

        # Clean up images input
        if isinstance(images, (Path, str)):
            images = [images]
        images = [get_location(image, cache_dir=output_dir/"downloads") for image in images]
        console.print(f"Processing {len(images)} image(s)")

        # Sheet-Components predictions
        component_files = self.sheet_component_detect(images, output_dir=output_dir)

        ocr_data = {}
        
        # Institutional Label Field Detection Model Predictions
        for stub, components in component_files.items():
            for component in components:
                if component.name.endswith("institutional_label.jpg"):
                    ocr_data[str(component)] = self.institutional_label_detect(
                        component, 
                        stub=stub,
                        output_dir=output_dir / stub,
                    )

        df = ocr_data_df(ocr_data, output_path=output_dir/"ocr_results.csv")

        # Write report
        if report:
            if len(images) == 1:
                # If report is for a single image, include in report name
                report_path = output_dir/f"hespi-report-{images[0].name}.html"
            else:
                # If report is for multiple images, make sure it is unique
                report_path = output_dir/"hespi-report.html"
                index = 1
                while report_path.exists():
                    index += 1
                    report_path = output_dir/f"hespi-report-{index}.html"
            
            write_report(report_path, component_files, df)

        return df

    def institutional_label_classify(self, component:Path, classification_csv:Path) -> str:
        console.print(f"Classifying institution label: '{component}'")
        console.print(f"Saving result to '{classification_csv}'")
        classifier_results = self.institutional_label_classifier(
            items=component,
            pretrained=self.institutional_label_classifier.pretrained,
            gpu=self.gpu,
            output_csv=classification_csv,
            verbose=False,
        )
        # Get classification of institution label
        if isinstance(classifier_results, pd.DataFrame):
            classification = classifier_results.iloc[0]["prediction"]
            console.print(
                f"'{component}' classified as '[red]{classification}[/red]'."
            )
        else:
            console.print(
                f"Could not get classification of institutional label '{component}'"
            )
            classification = ""

        return classification

    def institutional_label_detect(self, component, stub, output_dir) -> Dict:
        detection_results = {"id": stub}

        # Institutional Label Classification
        classification = self.institutional_label_classify(
            component=component,
            classification_csv = component.parent / f"label_classification.txt",
        )
        detection_results["label_classification"] = classification

        field_files = self.label_field_model_detect(
            [component],
            output_dir=output_dir,
        )

        institutional_label_stub = get_stub(component)
        institutional_label_dir = component.parent/institutional_label_stub
        predictions_path = institutional_label_dir/predictions_filename(institutional_label_stub)
        detection_results["predictions"] = predictions_path

        # Text Recognition on bounding boxes found by YOLO
        for fields in field_files.values():
            for field_file in fields:
                field_results = self.read_field_file(
                    field_file, 
                    classification,
                )
                detection_results.update(field_results)
        
        return detection_results

    def read_field_file(
        self,
        field_file:Path,
        classification:str = None,
    ) -> Dict:
        """Reads the text of an image of a field from an institutional label.

        Args:
            field_file (Path): The path to an image file of a text field. 
                The type of field needs to be in the second last component of the filename when split by periods.
                e.g. XXXXX.FIELD_NAME.jpg
                           ^^^^^^^^^^
            classification (str, optional): The classification of the institutional label from the institutional_label_classifier model.
                If it is 'printed' or 'typewriter' then the Tesseract OCR result will be favoured.
                Otherwise the TrOCR model result will be favoured.

        Returns:
            Dict: A dictionary with the results from TrOCR (if the `htr` flag is on), Tesseract and an adjusted form of the text,
                which changes the case depending on the field type and fuzzy matched with the reference database if `fuzzy` is requested.
        """
        field_file = Path(field_file)
        console.print("field_file:", field_file)
        field_file_components = field_file.name.split(".")
        assert len(field_file_components) >= 2
        field = field_file_components[-2]
        classification = classification or ""

        detection_results = {}        

            
        detection_results[f"{field}_image"] = field_file

        # HTR
        recognised_text = ""
        if self.htr:
            htr_text = self.trocr.get_text(field_file)
            if htr_text:
                print(f"HTR: {htr_text}")
                console.print(
                    f"Handwritten Text Recognition (TrOCR): [red]'{htr_text}'[/red]"
                )

                detection_results[f"{field}_TrOCR"] = htr_text
                recognised_text = htr_text

        # OCR
        tesseract_text = self.tesseract.get_text(field_file)
        if tesseract_text:
            detection_results[f"{field}_Tesseract"] = tesseract_text
            console.print(
                f"Optical Character Recognition (Tesseract): [red]'{tesseract_text}'[/red]"
            )

            if classification in ["printed", "typewriter"] or not recognised_text:
                recognised_text = tesseract_text

        # Adjust text if necessary
        if recognised_text:
            detection_results[field] = adjust_text(
                field, 
                recognised_text, 
                fuzzy=self.fuzzy, 
                fuzzy_cutoff=self.fuzzy_cutoff, 
                reference=self.reference,
            )

        return detection_results




from pathlib import Path
from typing import Dict
import pandas as pd
from rich.console import Console
from difflib import get_close_matches

console = Console()

DATA_DIR = Path(__file__).parent / "data"

label_fields = [
    "family",
    "genus",
    "species",
    "infrasp_taxon",
    "authority",
    "collector_number",
    "collector",
    "locality",
    "geolocation",
    "year",
    "month",
    "day",
]

def adjust_case(field, value):
    if field in ["genus", "family"]:
        return value.title()
    elif field == "species":
        return value.lower()
    
    return value


def read_reference(field):
    path = DATA_DIR / f"{field}.txt"
    if not path.exists():
        raise FileNotFoundError(f"No reference file for field '{field}'.")
    return path.read_text().strip().split("\n")


def ocr_data_df(data: dict, output_path: Path=None) -> pd.DataFrame:
    """    
    Creates a DataFrame of data, sorts columns and outputs a CSV with OCR values.

    Args:
        data (dict): The data coming from the text recognition models. 
            The keys are the institutional labels.
            The values are dictionaries with the fields as keys and the values as the recognised text.
        output_path (Path, optional): Where to save a CSV file if given. Defaults to None.

    Returns:
        pd.DataFrame: The text recognition data as a Pandas dataframe
    """
    df = pd.DataFrame.from_dict(data, orient="index")
    df = df.reset_index().rename(columns={"index": "institutional label"})
    
    # insert columns not included in dataframe, and re-order
    # including any columns not included in col_options to account for any updates
    col_options = [ "institutional label", "id" ] + label_fields

    missing_cols = [col for col in col_options if col not in df.columns]
    df[missing_cols] = ""
    extra_cols = [col for col in df.columns if col not in col_options]
    cols = col_options + extra_cols
    df = df[cols]
    df = df.fillna('')
    
    # CSV output
    if output_path:
        output_path = Path(output_path)
        output_path.parent.mkdir(exist_ok=True, parents=True)
        df.to_csv(output_path, index=False)

    return df


def adjust_text(field:str, recognised_text:str, fuzzy:bool, fuzzy_cutoff:float, reference:Dict):
    text_adjusted = adjust_case(field, recognised_text)

    # Match with database
    if fuzzy and field in reference:
        close_matches = get_close_matches(
            text_adjusted,
            reference[field],
            cutoff=fuzzy_cutoff,
            n=1,
        )
        if close_matches:
            text_adjusted = close_matches[0]

    if recognised_text != text_adjusted:
        console.print(
            f"Recognized text [red]'{recognised_text}'[/red] adjusted to [purple]'{text_adjusted}'[/purple]"
        )    
    return text_adjusted


def get_stub(path:Path) -> str:
    """
    Gets the part of the filename before the last extension

    Args:
        path (Path): The path to the file

    Returns:
        str: the part of the filename before the last extension
    """
    last_period = path.name.rfind(".")
    stub = path.name[:last_period] if last_period else path.name
    return stub

    


import typer
from rich.console import Console

from .hespi import DEFAULT_INSTITUTIONAL_LABEL_CLASSIFIER_WEIGHTS, DEFAULT_SHEET_COMPONENT_WEIGHTS, DEFAULT_LABEL_FIELD_WEIGHTS
from .download import get_location
from .util import DATA_DIR

console = Console()

app = typer.Typer()


@app.command()
def sheet_component_location():
    """ Shows the location of the default Sheet-Component model. """
    path = get_location(DEFAULT_SHEET_COMPONENT_WEIGHTS)
    console.print(f"The location of the default Sheet-Component model is:\n{path}")


@app.command()
def label_field_location():
    """ Shows the location of the default Label-Field model weights. """
    path = get_location(DEFAULT_LABEL_FIELD_WEIGHTS)
    console.print(f"The location of the default Label-Field model is:\n{path}")    


@app.command()
def bibtex():
    """ Shows the references in BibTeX format. """
    references = DATA_DIR / "references.bib"
    console.print(references.read_text())



import jinja2
from pathlib import Path
import pandas as pd
from typing import Dict

from .util import label_fields


def write_report(output:Path, component_files:Dict, ocr_df:pd.DataFrame):
    """
    Writes an HTML report of the detection found.

    Args:
        output (Path): The path to write the output HTML.
        component_files (Dict): A dictionary with all the files generated by running the sheet component model.
        ocr_df (pd.DataFrame): The character recognition data for the institutional labels.
    """
    template_dir = Path(__file__).parent/"templates"
    loader = jinja2.FileSystemLoader(template_dir)

    def relative_to_output(path):
        try:
            return Path(path).relative_to(output.parent)
        except Exception:
            return path

    def get_classification(path):
        return Path(path).name.split(".")[-2].replace("_", " ").title()

    def truncate(string):
        if len(string) < 30:
            return string
        return f"{string[:30]}..."

    env = jinja2.Environment(
        loader=loader,
        autoescape=jinja2.select_autoescape()
    )
    env.globals['relative_to_output'] = relative_to_output
    env.globals['get_classification'] = get_classification
    env.globals['len'] = len
    env.globals['truncate'] = truncate
    template = env.get_template("report-template.html")

    result = template.render(
        component_files=component_files,
        label_fields=label_fields,
        ocr_df=ocr_df,
    )
    with open(str(output), 'w') as f:
        print(f"Writing result to {output}")
        f.write(result)                




from pathlib import Path
from appdirs import user_cache_dir
from typing import Union
import hashlib
import urllib.request
import gzip, shutil

class DownloadError(Exception):
    pass


def get_cached_path(filename: str, cache_dir:Path=None) -> Path:
    """
    Returns a path in the ausdex directory in the user's cache.

    File may or may not exist.
    """
    cache_dir = cache_dir or Path(user_cache_dir("hespi"))
    cache_dir.mkdir(exist_ok=True, parents=True)
    return cache_dir / filename


def cached_download(url: str, local_path: Union[str, Path], force: bool = False) -> None:
    """
    Downloads a file if a local file does not already exist.

    Args:
        url (str): The url of the file to download.
        local_path (str, Path): The local path of where the file should be.
            If this file isn't there or the file size is zero then this function downloads it to this location.
        force (bool): Whether or not the file should be forced to download again even if present in the local path.
            Default False.

    Raises:
        DownloadError: Raises an exception if it cannot download the file.
        IOError: Raises an exception if the file does not exist or is empty after downloading.
    """
    local_path = Path(local_path)
    if (not local_path.exists() or local_path.stat().st_size == 0) or force:
        try:
            print(f"Downloading {url} to {local_path}")
            urllib.request.urlretrieve(url, local_path)
        except:
            raise DownloadError(f"Error downloading {url}")

    if not local_path.exists() or local_path.stat().st_size == 0:
        raise IOError(f"Error reading {local_path}")


def get_stem_extension(name):
    extension_location = name.rfind(".")
    if extension_location > 0:
        name_stem = name[:extension_location]
        extension = name[extension_location:]
    else:
        name_stem = name
        extension = ".dat"

    return name_stem, extension


def get_location(location:Union[str,Path], force:bool=False, cache_dir=None) -> Path:
    location = str(location)
    if location.startswith("http"):
        name_stem, extension = get_stem_extension(location.split("/")[-1])
        url_hash = hashlib.md5(location.encode()).hexdigest()

        # gunzip if necessary
        if extension == ".gz":
            name_stem, extension = get_stem_extension(name_stem)
            local_path_gz = get_cached_path(f"{name_stem}-{url_hash}{extension}.gz", cache_dir=cache_dir)
            local_path = get_cached_path(f"{name_stem}-{url_hash}{extension}", cache_dir=cache_dir)

            if not local_path.exists() or force:
                cached_download(location, local_path_gz, force=force)
                
                #https://stackoverflow.com/a/48466691
                with gzip.open(local_path_gz, 'r') as f_in, open(local_path, 'wb') as f_out:
                    shutil.copyfileobj(f_in, f_out)
                
                assert local_path.exists()
                local_path_gz.unlink()
        else:
            local_path = get_cached_path(f"{name_stem}-{url_hash}{extension}", cache_dir=cache_dir)
            cached_download(location, local_path, force=force)
    else:
        local_path = Path(location)
    
    if not local_path.exists() or local_path.stat().st_size == 0:
        raise IOError(f"Cannot read file {local_path}")

    return local_path



from typing import List
import typer
from pathlib import Path
import pandas as pd
from rich.console import Console

from .hespi import Hespi
from .hespi import DEFAULT_INSTITUTIONAL_LABEL_CLASSIFIER_WEIGHTS, DEFAULT_SHEET_COMPONENT_WEIGHTS, DEFAULT_LABEL_FIELD_WEIGHTS
from .ocr import TrOCRSize

console = Console()

app = typer.Typer()

@app.command()
def detect(
    images: List[str] = typer.Argument(..., help="A list of images to process."),
    output_dir: Path = typer.Option(
        ...,
        help="A directory to output the results.",
        prompt="Please specify a directory for the results",
    ),
    gpu: bool = typer.Option(True, help="Whether or not to use a GPU if available."),
    fuzzy: bool = typer.Option(
        True, help="Whether or not to use fuzzy matching from teh reference database."
    ),
    fuzzy_cutoff: float = typer.Option(
        0.8, min=0.0, max=1.0, help="The threshold for the fuzzy matching score to use."
    ),
    htr: bool = typer.Option(
        True,
        help="Whether or not to do handwritten text recognition using Microsoft's TrOCR.",
    ),
    trocr_size: TrOCRSize = typer.Option(
        TrOCRSize.BASE.value,
        help="The size of the TrOCR model to use for handwritten text recognition.",
        case_sensitive=False,
    ),
    sheet_component_weights: str = typer.Option(
        DEFAULT_SHEET_COMPONENT_WEIGHTS, 
        help="The path to the sheet component model weights.",
        envvar="HESPI_SHEET_COMPONENT_WEIGHTS",
    ),
    label_field_weights: str = typer.Option(
        DEFAULT_LABEL_FIELD_WEIGHTS,
        help="The path to the Label-Field model weights.",
        envvar="HESPI_LABEL_FIELD_WEIGHTS",
    ),
    institutional_label_classifier_weights: str = typer.Option(
        DEFAULT_INSTITUTIONAL_LABEL_CLASSIFIER_WEIGHTS,
        help="The path to the institutional label classifier weights.",
    ),
    force_download:bool = typer.Option(False, help="Whether or not to force download model weights even if a weights file is present."),
    tmp_dir:str = None,
    batch_size:int = typer.Option(4, min=1, help="The maximum batch size from run the sheet component model."),
    sheet_component_res:int = typer.Option(1280, min=640, help="The resolution of images to use for the Sheet-Component model."),
    label_field_res:int = typer.Option(1280, min=640, help="The resolution of images to use for the Label-Field model."),
) -> pd.DataFrame:
    """
    HErbarium Specimen sheet PIpeline

    Takes a herbarium specimen sheet image detects components such as the institutional label, swatch, etc.
    It then classifies whether the institutional label is printed, typed, handwritten or a combination.
    If then detects the fields of the institutional label and attempts to read them through OCR and HTR models.
    """
    hespi = Hespi(
        trocr_size=trocr_size,
        sheet_component_weights=sheet_component_weights,
        label_field_weights=label_field_weights,
        institutional_label_classifier_weights=institutional_label_classifier_weights,
        force_download=force_download,
        gpu=gpu,
        fuzzy=fuzzy,
        fuzzy_cutoff=fuzzy_cutoff,
        htr=htr,
        tmp_dir=tmp_dir,
        batch_size=batch_size,
        sheet_component_res=sheet_component_res,
        label_field_res=label_field_res,
    )
    return hespi.detect(images, output_dir)
    


import tempfile
from shutil import move
from pathlib import Path
from PIL import Image
from collections import defaultdict
from rich.console import Console


try:
    from ultralytics.models.yolo.detect.predict import DetectionPredictor
except ImportError:
    from ultralytics.yolo.v8.detect.predict import DetectionPredictor


console = Console()

from .util import get_stub

def predictions_filename(stub):
    return f"{stub}-predictions.jpg"


def yolo_output(model, images, output_dir, tmp_dir_prefix=None, batch_size=4, res=1280):
    if not model.predictor:
        model.predictor = DetectionPredictor()
        model.predictor.setup_model(model=model.model)

    tmp_dir = tempfile.TemporaryDirectory(prefix=tmp_dir_prefix)
    tmp_dir_path = Path(tmp_dir.name)
    console.print(f"Using temporary directory '{tmp_dir_path}'")

    model.predictor.save_dir = tmp_dir_path

    results = model.predict(source=images, show=False, save=True, batch_size=batch_size, imgsz=res)
    output_files = defaultdict(list)

    output_dir = Path(output_dir)

    for index, (image, predictions) in enumerate(zip(images, results)):
        stub = get_stub(image)
        image_output_dir = output_dir / stub
        image_output_dir.mkdir(exist_ok=True, parents=True)
        prediction_path = image_output_dir / predictions_filename(stub)

        prediction_path_tmp_location = Path(model.predictor.save_dir)/f"image{index}.jpg"
        assert prediction_path_tmp_location.exists()

        console.print(
            f"Saving sheet component predicitons with bounding boxes to '{prediction_path}'"
        )
        move(prediction_path_tmp_location, prediction_path)

        for prediction_index, boxes in enumerate(predictions.boxes.boxes):
            category_index = int(boxes[5].int().cpu().numpy())
            category = (
                model.names[category_index].replace(" ", "_").replace(":", "").strip()
            )

            x0, y0, x1, y1 = boxes[:4].cpu().numpy()

            # open image
            im = Image.open(image)
            im_crop = im.crop((x0, y0, x1, y1))
            output_path = (
                image_output_dir / f"{stub}.{prediction_index}.{category}.jpg"
            )
            console.print(f"Saving {category} to '{output_path}'")
            im_crop.convert('RGB').save(output_path)
            output_files[stub].append(output_path)

    tmp_dir.cleanup()

    return output_files